STARTUP_TIMEOUT: float
"""How long to wait for the render server to start in seconds."""

RENDER_TIMEOUT: float
"""Timeout per rendering request in seconds."""

NODEJS_BINARY: str
"""nodejs binary to run javascript."""
